'''
Synthetic instance generator for Materials Discovery
Copyright 2014, Institute for Computational Sustainability
See README for license and citation information.  

Main contributors: 
Ronan Le Bras (Cornell University Dept. of Computer Science) lebras@cs.cornell.edu
Richard Bernstein (Cornell University Dept. of Computer Science) rab38@cornell.edu
John Gregoire (Caltech JCAP Energy Innovation Hub) gregoire@caltech.edu

SyntheticXRD.py - helper function for calculating XRD patterns, I/O, etc
Modified 11/06/2014
'''

import numpy
import operator
import copy
import random
import bisect
import string

INSTANCE_FORMAT_VER="1.0"

def peakshape(q, pk):
    """
    Calculate the signal at the given q values for a gaussian peak with the given (mean, height, stdev) 
    """
    return pk[1]*numpy.exp(-2.0*(q-pk[0])**2/pk[2]**2)

def q_qgrid_ind(qgrid, index='all'):
    """
    Get a vector of q values from the given [start,step,count] instructions
    """
    if index=='all':
        index=numpy.array(range(numpy.uint16(qgrid[2])), dtype=numpy.float32)
    elif isinstance(index, list):
        index=numpy.array(index)
    return qgrid[0]+qgrid[1]*index

def calcxrd(phases, regions, export_comp, export_region, qgrid, mergepeakwidthfrac=0):
    """
    Generate all the peak, phase and region information based on the given model information
    """
    synthpeaks=[]
    synthpatterns=[]
    synthphaseconc=[]
    synthmergedpeaks=[]
    qvals=numpy.float32(q_qgrid_ind(qgrid))
    for excomp, exregion in zip(export_comp, export_region):
        pklist=[]
        phconc=[]
        pattern=numpy.zeros(qvals.shape, dtype='float32')
        for (phasefuncs, ph) in zip(regions[exregion], phases):
            pks_ph=[[v*f(excomp) for v, f in zip(pk, phasefuncs)] for pk in ph]
            pklist+=pks_ph
        for pk in pklist:
            pattern+=peakshape(qvals, pk)
        synthpatterns+=[pattern]
        synthphaseconc+=[[regions[exregion][i][1](excomp)*regions[exregion][i][2](excomp) for i in xrange(len(phases))]]
        pklist = [pk for pk in pklist if pk[1]>0.]
        peakorder = sorted(xrange(len(pklist)),key=[pk[0] for pk in pklist].__getitem__)
        pklist = [pklist[x] for x in peakorder]

        mergedpeaks=[]
        if mergepeakwidthfrac>0:
            pkar=numpy.float32(pklist)
            minsep=numpy.min(2.*(pkar[1:, 0]-pkar[:-1, 0])/(pkar[1:, 2]+pkar[:-1, 2]))
            while minsep<mergepeakwidthfrac:
                lowind=numpy.argmin(2.*(pkar[1:, 0]-pkar[:-1, 0])/(pkar[1:, 2]+pkar[:-1, 2]))
                pk1=pklist.pop(lowind)
                pk2=pklist.pop(lowind)#the list is one shorter befre this executes
                newpk=[(pk1[0]*pk1[1]+pk2[0]*pk2[1])/(pk1[1]+pk2[1]), max(pk1[1], pk2[1]), (pk1[2]*pk1[1]+pk2[2]*pk2[1])/(pk1[1]+pk2[1])+pk2[0]-pk1[0]]
                pklist+=[newpk]
                pklist.sort(key=operator.itemgetter(0))
                mergedpeaks+=[[newpk, pk1, pk2]]
                pkar=numpy.float32(pklist)
                minsep=numpy.min(2.*(pkar[1:, 0]-pkar[:-1, 0])/(pkar[1:, 2]+pkar[:-1, 2]))
        synthmergedpeaks+=[mergedpeaks]
        synthpeaks+=[pklist]

    synthphaseconc=numpy.float32(synthphaseconc)

    return synthpeaks, synthpatterns, synthphaseconc, synthmergedpeaks

def saveInstance(folder, name, elements, synthpatterns, synthpeaks, synthphases, export_comp, qgrid, description, modeluuid,intensityscale):
    """
    Save information about the instance to a .txt data file.
    """

    qvals=numpy.float32(q_qgrid_ind(qgrid))
    nsamples = len(synthpatterns)
    compdomainlabels = ["C"+e for e in elements]
    instfilename=folder+'/'+name+'_inst.txt'
    finst = open(instfilename,"w")
    finst.write("Description="+description+"\n")
    finst.write("UUID="+modeluuid+"\n")
    finst.write("Format_Version="+INSTANCE_FORMAT_VER+"\n")
    finst.write("\n")

    finst.write("// Metadata\nM={0:d}\nElements={1:s}\n".format(len(elements),string.join(elements,",")))
    finst.write("N={0:d}\n".format(nsamples))
#    finst.write("Deposition=\n")
    finst.write("Composition={0:s}\n".format(string.join(compdomainlabels,",")))
    finst.write("\n")
    
    finst.write("// Deposition/Composition data\n")
    for i in xrange(len(export_comp[0])):
        finst.write("{0:s}={1:s}\n".format(compdomainlabels[i],string.join(["{0:.3f}".format(c[i]) for c in export_comp],",")))
    finst.write("\n")
    finst.write("//Integrated counts data\n")
    finst.write("Q={0:s}\n".format(string.join(["{0:.3f}".format(q) for q in qvals],",")))
    for i in xrange(nsamples):
        finst.write("I{0:d}={1:s}\n".format(i+1,string.join(["{0:.6f}".format(y) for y in synthpatterns[i]],",")))
    finst.write("\n")
    for i in xrange(nsamples):
        finst.write("P{0:d}={1:s}\n".format(i+1,string.join(["[{0:.3f},{1:.3f},{2:.3f}]".format(p[0],p[1],p[2]) for p in synthpeaks[i]],",")))
    finst.close()    
        
    return instfilename

def saveSolution(folder, name, synthphases, regions, phases, export_comp, qgrid, qbasisoffsets, description, modeluuid,intensityscale):
    """
    Save information about the correct solution to a .txt data file.
    """
    qvals=numpy.float32(q_qgrid_ind(qgrid))
    nsamples = len(regions)
    nphases = len(phases)
    
    solfilename=folder+'/'+name+'_sol.txt'
    fsol = open(solfilename,"w")
    fsol.write("Description="+description+"\n")
    fsol.write("UUID="+modeluuid+"\n")
    fsol.write("Format_Version="+INSTANCE_FORMAT_VER+"\n")
    fsol.write("\n")
    
    fsol.write("// Number of phases\n")
    fsol.write("K={0:d}\n".format(len(phases)))
    fsol.write("Params=[R,C],[Q,B,C,S],[P,C,S,W]\n")
    fsol.write("\n")
    fsol.write("// Phase patterns (basis)\n")
    fsol.write("Q={0:s}\n".format(string.join(["{0:.3f}".format(q) for q in qvals],",")))
    
    pattern=numpy.zeros(qvals.shape, dtype='float32')
    for i in xrange(nphases):
        pattern[:]=0.
        for pk in phases[i]:
            if pk[1]>0.:
                pattern+=peakshape(qvals, (pk[0]*qbasisoffsets[i],pk[1],pk[2]))
        fsol.write("B{0:d}={1:s}\n".format(i+1,string.join(["{0:.6f}".format(y) for y in pattern],",")))
    fsol.write("\n")
    fsol.write("// Phase peaks (basis as peaks)\n")
    for i in xrange(nphases):
        fsol.write("P{0:d}={1:s}\n".format(i+1,string.join(["[{0:.3f},{1:.3f},{2:.3f}]".format(p[0]*qbasisoffsets[i],p[1],p[2]) for p in phases[i]],",")))
    fsol.write("\n")

    fsol.write("// Per-phase model for each sample\n")
    for j in xrange(nsamples):
        for i in xrange(nphases):
            pattern[:]=0.
            for pk in phases[i]:
                if pk[1]>0.:
                    pattern+=peakshape(qvals, (pk[0]*regions[j][i][0](export_comp[j]),pk[1]*regions[j][i][1](export_comp[j]),pk[2]*regions[j][i][2](export_comp[j])))
            fsol.write("R{0:d}_{1:d}={2:s}\n".format(j+1,i+1,string.join(["{0:.6f}".format(y) for y in pattern],",")))
    fsol.write("\n")
    
    fsol.write("// Phase concentrations at each sample\n")
    for i in xrange(nsamples):
        fsol.write("C{0:d}={1:s}\n".format(i+1,string.join(["{0:.6f}".format(c) for c in synthphases[i]],",")))
    fsol.write("\n")
    fsol.write("// Phase shifts at each sample\n")
    for i in xrange(nsamples):
        fsol.write("S{0:d}={1:s}\n".format(i+1,string.join(["{0:.6f}".format(regions[i][p][0](export_comp[i])/qbasisoffsets[p] if synthphases[i][p]>0. else 0.) for p in xrange(nphases)],",")))
    fsol.write("\n")
    fsol.write("// Peak width multiplier at each sample\n")
    for i in xrange(nsamples):
        fsol.write("W{0:d}={1:s}\n".format(i+1,string.join(["{0:.6f}".format(regions[i][p][2](export_comp[i]) if synthphases[i][p]>0. else 0.) for p in xrange(nphases)],",")))
    fsol.close()        
    return solfilename

def removepeaks(synthpeaks, num, qgrid):
    indeces=[]
    weights=[]
    for count, peaks in enumerate(synthpeaks):
        for count2, pk in enumerate(peaks):
            indeces+=[[count, count2]]
            weights+=[1./pk[1]**2.]
    weights=numpy.float64(weights)
    removedpointpk=[]
    if num>len(indeces):
        num=len(indeces)
        print 'Too many peaks specified for removal - removing all peaks instead'
    while num>0:
        cumWeights = numpy.cumsum(weights)
        choiceVal = random.random()*cumWeights[-1]
        ri = bisect.bisect(cumWeights, choiceVal)
        weights[ri]=0.
        print 'Peak Removed ', (indeces[ri], synthpeaks[indeces[ri][0]][indeces[ri][1]])
        removedpointpk+=[(indeces[ri], copy.copy(synthpeaks[indeces[ri][0]][indeces[ri][1]]))]
        synthpeaks[indeces[ri][0]][indeces[ri][1]][1]=0.

        num -= 1
    removedpointpk.sort(key=operator.itemgetter(0,0))
    synthpeaks = [[pk for pk in pklist if pk[1]>0.] for pklist in synthpeaks]

    #now recalc patterns
    qvals=numpy.float32(q_qgrid_ind(qgrid))
    synthpatterns=[]
    for pklist in synthpeaks:
        pattern=numpy.zeros(qvals.shape, dtype='float32')
        for pk in pklist:
            pattern+=peakshape(qvals, pk)
        synthpatterns+=[pattern]
    return synthpeaks, synthpatterns, removedpointpk

def compgrid(compsize=2, ternary=True):
    """
    Generate a regular grid of composition mixtures using the given composition ranges and spacing.
    """
    stepsize = 1./int(compsize-1)
    if ternary:
        return [[c*stepsize,d*stepsize,1.-stepsize*(c+d)] for c in xrange(compsize) for d in xrange(compsize-c)] 
    else:
        return [[c*stepsize,1.-c*stepsize] for c in xrange(compsize)]
